@@ -72,6 +72,22 @@ static char *order[6] = {
     "vertical bgr",
     "no subpixels"};
 
+static const struct {
+    char	    *string;
+    unsigned long   flag;
+} mode_flags[] = {
+    { "+HSync", RR_HSyncPositive },
+    { "-HSync", RR_HSyncNegative },
+    { "+VSync", RR_VSyncPositive },
+    { "-VSync", RR_VSyncNegative },
+    { "Interlace", RR_Interlace },
+    { "DoubleScan", RR_DoubleScan },
+    { "CSync",	    RR_CSync },
+    { "+CSync",	    RR_CSyncPositive },
+    { "-CSync",	    RR_CSyncNegative },
+    { NULL,	    0 }
+};
+
 static void
 usage(void)
 {
@@ -113,15 +129,18 @@ usage(void)
     fprintf(stderr, "      --below <output>\n");
     fprintf(stderr, "      --same-as <output>\n");
     fprintf(stderr, "      --set <property> <value>\n");
+    fprintf(stderr, "      --scale <x>x<y>\n");
+    fprintf(stderr, "      --transform <a>,<b>,<c>,<d>,<e>,<f>,<g>,<h>,<i>\n");
     fprintf(stderr, "      --off\n");
     fprintf(stderr, "      --crtc <crtc>\n");
+    fprintf(stderr, "      --panning <w>x<h>[+<x>+<y>[/<track:w>x<h>+<x>+<y>[/<border:l>/<t>/<r>/<b>]]]\n");
     fprintf(stderr, "  --newmode <name> <clock MHz>\n");
     fprintf(stderr, "            <hdisp> <hsync-start> <hsync-end> <htotal>\n");
     fprintf(stderr, "            <vdisp> <vsync-start> <vsync-end> <vtotal>\n");
     fprintf(stderr, "            [+HSync] [-HSync] [+VSync] [-VSync]\n");
     fprintf(stderr, "  --rmmode <name>\n");
-    fprintf(stderr, "  --addmode <output> <mode>\n");
-    fprintf(stderr, "  --delmode <output> <mode>\n");
+    fprintf(stderr, "  --addmode <output> <name>\n");
+    fprintf(stderr, "  --delmode <output> <name>\n");
 #endif
 
     exit(1);
@@ -141,6 +160,17 @@ fatal (const char *format, ...)
     /*NOTREACHED*/
 }
 
+static void
+warning (const char *format, ...)
+{
+    va_list ap;
+    
+    va_start (ap, format);
+    fprintf (stderr, "%s: ", program_name);
+    vfprintf (stderr, format, ap);
+    va_end (ap);
+}
+
 static char *
 rotation_name (Rotation rotation)
 {
@@ -180,6 +210,18 @@ typedef enum _relation {
     left_of, right_of, above, below, same_as,
 } relation_t;
 
+typedef struct {
+    int	    x, y, width, height;
+} rectangle_t;
+
+typedef struct {
+    int	    x1, y1, x2, y2;
+} box_t;
+
+typedef struct {
+    int	    x, y;
+} point_t;
+
 typedef enum _changes {
     changes_none = 0,
     changes_crtc = (1 << 0),
@@ -191,6 +233,8 @@ typedef enum _changes {
     changes_automatic = (1 << 6),
     changes_refresh = (1 << 7),
     changes_property = (1 << 8),
+    changes_transform = (1 << 9),
+    changes_panning = (1 << 10),
 } changes_t;
 
 typedef enum _name_kind {
@@ -210,20 +254,30 @@ typedef struct {
 
 typedef struct _crtc crtc_t;
 typedef struct _output	output_t;
+typedef struct _transform transform_t;
 typedef struct _umode	umode_t;
 typedef struct _output_prop output_prop_t;
 
+struct _transform {
+    XTransform	    transform;
+    char	    *filter;
+    int		    nparams;
+    XFixed	    *params;
+};
+
 struct _crtc {
     name_t	    crtc;
     Bool	    changing;
     XRRCrtcInfo	    *crtc_info;
 
     XRRModeInfo	    *mode_info;
+    XRRPanning      *panning_info;
     int		    x;
     int		    y;
     Rotation	    rotation;
     output_t	    **outputs;
     int		    noutput;
+    transform_t	    current_transform, pending_transform;
 };
 
 struct _output_prop {
@@ -244,6 +298,7 @@ struct _output {
     
     name_t	    crtc;
     crtc_t	    *crtc_info;
+    crtc_t	    *current_crtc_info;
     
     name_t	    mode;
     float	    refresh;
@@ -257,7 +312,10 @@ struct _output {
     int		    x, y;
     Rotation	    rotation;
     
+    XRRPanning      panning;
+
     Bool    	    automatic;
+    transform_t	    transform;
 };
 
 typedef enum _umode_action {
@@ -305,6 +363,7 @@ static char	*dpi_output = NULL;
 static Bool	dryrun = False;
 static int	minWidth, maxWidth, minHeight, maxHeight;
 static Bool    	has_1_2 = False;
+static Bool    	has_1_3 = False;
 
 static int
 mode_height (XRRModeInfo *mode_info, Rotation rotation)
@@ -336,6 +395,81 @@ mode_width (XRRModeInfo *mode_info, Rotation rotation)
     }
 }
 
+Bool
+transform_point (XTransform *transform, double *xp, double *yp)
+{
+    double  vector[3];
+    double  result[3];
+    int	    i, j;
+    double  partial, v;
+
+    vector[0] = *xp;
+    vector[1] = *yp;
+    vector[2] = 1;
+    for (j = 0; j < 3; j++)
+    {
+	v = 0;
+	for (i = 0; i < 3; i++)
+	    v += (XFixedToDouble (transform->matrix[j][i]) * vector[i]);
+	if (v > 32767 || v < -32767)
+	    return False;
+	result[j] = v;
+    }
+    if (!result[2])
+	return False;
+    for (j = 0; j < 2; j++)
+	vector[j] = result[j] / result[2];
+    *xp = vector[0];
+    *yp = vector[1];
+    return True;
+}
+
+Bool
+path_bounds (XTransform *transform, point_t *points, int npoints, box_t *box)
+{
+    int	    i;
+    box_t   point;
+
+    for (i = 0; i < npoints; i++) {
+	double	x, y;
+	x = points[i].x;
+	y = points[i].y;
+	transform_point (transform, &x, &y);
+	point.x1 = floor (x);
+	point.y1 = floor (y);
+	point.x2 = ceil (x);
+	point.y2 = ceil (y);
+	if (i == 0)
+	    *box = point;
+	else {
+	    if (point.x1 < box->x1) box->x1 = point.x1;
+	    if (point.y1 < box->y1) box->y1 = point.y1;
+	    if (point.x2 > box->x2) box->x2 = point.x2;
+	    if (point.y2 > box->y2) box->y2 = point.y2;
+	}
+    }
+}
+
+static void
+mode_geometry (XRRModeInfo *mode_info, Rotation rotation,
+	       XTransform *transform,
+	       box_t *bounds)
+{
+    point_t rect[4];
+    int	width = mode_width (mode_info, rotation);
+    int height = mode_height (mode_info, rotation);
+
+    rect[0].x = 0;
+    rect[0].y = 0;
+    rect[1].x = width;
+    rect[1].y = 0;
+    rect[2].x = width;
+    rect[2].y = height;
+    rect[3].x = 0;
+    rect[3].y = height;
+    path_bounds (transform, rect, 4, bounds);
+}
+
 /* v refresh frequency in Hz */
 static float
 mode_refresh (XRRModeInfo *mode_info)
@@ -411,7 +545,8 @@ set_name_all (name_t *name, name_t *old)
 static void
 set_name (name_t *name, char *string, name_kind_t valid)
 {
-    XID	xid;
+    unsigned int xid; /* don't make it XID (which is unsigned long):
+			 scanf() takes unsigned int */
     int index;
 
     if ((valid & name_xid) && sscanf (string, "0x%x", &xid) == 1)
@@ -424,6 +559,53 @@ set_name (name_t *name, char *string, name_kind_t valid)
 	usage ();
 }
 
+static void
+init_transform (transform_t *transform)
+{
+    int x;
+    memset (&transform->transform, '\0', sizeof (transform->transform));
+    for (x = 0; x < 3; x++)
+	transform->transform.matrix[x][x] = XDoubleToFixed (1.0);
+    transform->filter = "";
+    transform->nparams = 0;
+    transform->params = NULL;
+}
+
+static void
+set_transform (transform_t  *dest,
+	       XTransform   *transform,
+	       char	    *filter,
+	       XFixed	    *params,
+	       int	    nparams)
+{
+    dest->transform = *transform;
+    dest->filter = strdup (filter);
+    dest->nparams = nparams;
+    dest->params = malloc (nparams * sizeof (XFixed));
+    memcpy (dest->params, params, nparams * sizeof (XFixed));
+}
+
+static void
+copy_transform (transform_t *dest, transform_t *src)
+{
+    set_transform (dest, &src->transform,
+		   src->filter, src->params, src->nparams);
+}
+
+static Bool
+equal_transform (transform_t *a, transform_t *b)
+{
+    if (memcmp (&a->transform, &b->transform, sizeof (XTransform)) != 0)
+	return False;
+    if (strcmp (a->filter, b->filter) != 0)
+	return False;
+    if (a->nparams != b->nparams)
+	return False;
+    if (memcmp (a->params, b->params, a->nparams * sizeof (XFixed)) != 0)
+	return False;
+    return True;
+}
+
 static output_t *
 add_output (void)
 {
@@ -669,6 +851,17 @@ crtc_can_use_rotation (crtc_t *crtc, Rotation rotation)
     return False;
 }
 
+static Bool
+crtc_can_use_transform (crtc_t *crtc, XTransform *transform)
+{
+    int	major, minor;
+
+    XRRQueryVersion (dpy, &major, &minor);
+    if (major > 1 || (major == 1 && minor >= 3))
+	return True;
+    return False;
+}
+
 /*
  * Report only rotations that are supported by all crtcs
  */
@@ -720,8 +913,8 @@ set_output_info (output_t *output, RROutput xid, XRROutputInfo *output_info)
 {
     /* sanity check output info */
     if (output_info->connection != RR_Disconnected && !output_info->nmode)
-	fatal ("Output %s is not disconnected but has no modes\n",
-	       output_info->name);
+	warning ("Output %s is not disconnected but has no modes\n",
+		 output_info->name);
     
     /* set output name and info */
     if (!(output->output.kind & name_xid))
@@ -826,6 +1019,15 @@ set_output_info (output_t *output, RROutput xid, XRROutputInfo *output_info)
 	       output->output.string,
 	       rotation_name (output->rotation),
 	       reflection_name (output->rotation));
+
+    /* set transformation */
+    if (!(output->changes & changes_transform))
+    {
+	if (output->crtc_info)
+	    copy_transform (&output->transform, &output->crtc_info->current_transform);
+	else
+	    init_transform (&output->transform);
+    }
 }
     
 static void
@@ -853,10 +1055,20 @@ get_crtcs (void)
     for (c = 0; c < res->ncrtc; c++)
     {
 	XRRCrtcInfo *crtc_info = XRRGetCrtcInfo (dpy, res, res->crtcs[c]);
+#if RANDR_MAJOR > 1 || RANDR_MINOR >= 3
+	XRRCrtcTransformAttributes  *attr;
+#endif
+	int	    x;
+	XRRPanning  *panning_info = NULL;
+
+	if (has_1_3)
+	    panning_info = XRRGetPanning  (dpy, res, res->crtcs[c]);
+
 	set_name_xid (&crtcs[c].crtc, res->crtcs[c]);
 	set_name_index (&crtcs[c].crtc, c);
 	if (!crtc_info) fatal ("could not get crtc 0x%x information", res->crtcs[c]);
 	crtcs[c].crtc_info = crtc_info;
+	crtcs[c].panning_info = panning_info;
 	if (crtc_info->mode == None)
 	{
 	    crtcs[c].mode_info = NULL;
@@ -864,7 +1076,22 @@ get_crtcs (void)
 	    crtcs[c].y = 0;
 	    crtcs[c].rotation = RR_Rotate_0;
 	}
-    }
+#if RANDR_MAJOR > 1 || RANDR_MINOR >= 3
+	if (XRRGetCrtcTransform (dpy, res->crtcs[c], &attr) && attr) {
+	    set_transform (&crtcs[c].current_transform,
+			   &attr->currentTransform,
+			   attr->currentFilter,
+			   attr->currentParams,
+			   attr->currentNparams);
+	    XFree (attr);
+	}
+	else
+#endif
+	{
+	    init_transform (&crtcs[c].current_transform);
+	}
+	copy_transform (&crtcs[c].pending_transform, &crtcs[c].current_transform);
+   }
 }
 
 static void
@@ -879,7 +1106,8 @@ crtc_add_output (crtc_t *crtc, output_t *output)
 	crtc->y = output->y;
 	crtc->rotation = output->rotation;
 	crtc->mode_info = output->mode_info;
-    }
+	copy_transform (&crtc->pending_transform, &output->transform);
+   }
     if (!crtc->outputs) fatal ("out of memory");
     crtc->outputs[crtc->noutput++] = output;
 }
@@ -896,6 +1124,24 @@ set_crtcs (void)
     }
 }
 
+static void
+set_panning (void)
+{
+    output_t	*output;
+
+    for (output = outputs; output; output = output->next)
+    {
+	if (! output->crtc_info)
+	    continue;
+	if (! (output->changes & changes_panning))
+	    continue;
+	if (! output->crtc_info->panning_info)
+	    output->crtc_info->panning_info = malloc (sizeof(XRRPanning));
+	memcpy (output->crtc_info->panning_info, &output->panning, sizeof(XRRPanning));
+	output->crtc_info->changing = 1;
+    }
+}
+
 static Status
 crtc_disable (crtc_t *crtc)
 {
@@ -908,6 +1154,20 @@ crtc_disable (crtc_t *crtc)
 			     0, 0, None, RR_Rotate_0, NULL, 0);
 }
 
+static void
+crtc_set_transform (crtc_t *crtc, transform_t *transform)
+{
+    int	major, minor;
+
+    XRRQueryVersion (dpy, &major, &minor);
+    if (major > 1 || (major == 1 && minor >= 3))
+	XRRSetCrtcTransform (dpy, crtc->crtc.xid,
+			     &transform->transform,
+			     transform->filter,
+			     transform->params,
+			     transform->nparams);
+}
+
 static Status
 crtc_revert (crtc_t *crtc)
 {
@@ -918,6 +1178,9 @@ crtc_revert (crtc_t *crtc)
 	
     if (dryrun)
 	return RRSetConfigSuccess;
+
+    if (!equal_transform (&crtc->current_transform, &crtc->pending_transform))
+	crtc_set_transform (crtc, &crtc->current_transform);
     return XRRSetCrtcConfig (dpy, res, crtc->crtc.xid, CurrentTime,
 			    crtc_info->x, crtc_info->y,
 			    crtc_info->mode, crtc_info->rotation,
@@ -953,9 +1216,19 @@ crtc_apply (crtc_t *crtc)
     if (dryrun)
 	s = RRSetConfigSuccess;
     else
+    {
+	if (!equal_transform (&crtc->current_transform, &crtc->pending_transform))
+	    crtc_set_transform (crtc, &crtc->pending_transform);
 	s = XRRSetCrtcConfig (dpy, res, crtc->crtc.xid, CurrentTime,
 			      crtc->x, crtc->y, mode, crtc->rotation,
 			      rr_outputs, crtc->noutput);
+	if (s == RRSetConfigSuccess && crtc->panning_info) {
+	    if (has_1_3)
+		s = XRRSetPanning (dpy, res, crtc->crtc.xid, crtc->panning_info);
+	    else
+		fatal ("panning needs RandR 1.3");
+	}
+    }
     free (rr_outputs);
     return s;
 }
@@ -1061,16 +1334,21 @@ apply (void)
 	{
 	    XRRModeInfo	*old_mode = find_mode_by_xid (crtc_info->mode);
 	    int x, y, w, h;
+	    box_t bounds;
 
 	    if (!old_mode) 
 		panic (RRSetConfigFailed, crtc);
 	    
 	    /* old position and size information */
-	    x = crtc_info->x;
-	    y = crtc_info->y;
-	    w = mode_width (old_mode, crtc_info->rotation);
-	    h = mode_height (old_mode, crtc_info->rotation);
-	    
+	    mode_geometry (old_mode, crtc_info->rotation,
+			   &crtc->current_transform.transform,
+			   &bounds);
+
+	    x = crtc_info->x + bounds.x1;
+	    y = crtc_info->y + bounds.y1;
+	    w = bounds.x2 - bounds.x1;
+	    h = bounds.y2 - bounds.y1;
+
 	    /* if it fits, skip it */
 	    if (x + w <= fb_width && y + h <= fb_height) 
 		continue;
@@ -1220,6 +1498,75 @@ mark_changing_crtcs (void)
     }
 }
 
+/*
+ * Test whether 'crtc' can be used for 'output'
+ */
+Bool
+check_crtc_for_output (crtc_t *crtc, output_t *output, Bool ignore_state)
+{
+    int		c;
+    int		l;
+    output_t    *other;
+    
+    for (c = 0; c < output->output_info->ncrtc; c++)
+	if (output->output_info->crtcs[c] == crtc->crtc.xid)
+	    break;
+    if (c == output->output_info->ncrtc)
+	return False;
+    for (other = outputs; other; other = other->next)
+    {
+	if (other == output)
+	    continue;
+
+	if (other->mode_info == NULL)
+	    continue;
+
+	if (other->crtc_info != crtc)
+	    continue;
+
+	/* see if the output connected to the crtc can clone to this output */
+	for (l = 0; l < output->output_info->nclone; l++)
+	    if (output->output_info->clones[l] == other->output.xid)
+		break;
+	/* not on the list, can't clone */
+	if (l == output->output_info->nclone) 
+	    return False;
+    }
+
+    if (ignore_state)
+	return True;
+
+    if (crtc->noutput)
+    {
+	/* make sure the state matches */
+	if (crtc->mode_info != output->mode_info)
+	    return False;
+	if (crtc->x != output->x)
+	    return False;
+	if (crtc->y != output->y)
+	    return False;
+	if (crtc->rotation != output->rotation)
+	    return False;
+	if (!equal_transform (&crtc->current_transform, &output->transform))
+	    return False;
+    }
+    else if (crtc->crtc_info->noutput)
+    {
+	/* make sure the state matches the already used state */
+	XRRModeInfo *mode = find_mode_by_xid (crtc->crtc_info->mode);
+
+	if (mode != output->mode_info)
+	    return False;
+	if (crtc->crtc_info->x != output->x)
+	    return False;
+	if (crtc->crtc_info->y != output->y)
+	    return False;
+	if (crtc->crtc_info->rotation != output->rotation)
+	    return False;
+    }
+    return True;
+}
+
 crtc_t *
 find_crtc_for_output (output_t *output)
 {
@@ -1228,42 +1575,12 @@ find_crtc_for_output (output_t *output)
     for (c = 0; c < output->output_info->ncrtc; c++)
     {
 	crtc_t	    *crtc;
-	int	    l;
-	output_t    *other;
 
 	crtc = find_crtc_by_xid (output->output_info->crtcs[c]);
 	if (!crtc) fatal ("cannot find crtc 0x%x\n", output->output_info->crtcs[c]);
 
-	for (other = outputs; other; other = other->next)
-	{
-	    if (other == output)
-		continue;
-
-	    if (other->mode_info == NULL)
-		continue;
-
-	    if (other->crtc_info != crtc)
-		continue;
-	    
-	    /* see if the output connected to the crtc can clone to this output */
-	    for (l = 0; l < output->output_info->nclone; l++)
-		if (output->output_info->clones[l] == other->output.xid)
-		    break;
-	    /* not on the list, can't clone */
-	    if (l == output->output_info->nclone) break;
-	}
-	if (other) continue;
-	
-
-	if (crtc->noutput)
-	{
-	    /* make sure the state matches */
-	    if (crtc->mode_info != output->mode_info) continue;
-	    if (crtc->x != output->x) continue;
-	    if (crtc->y != output->y) continue;
-	    if (crtc->rotation != output->rotation) continue;
-	}
-	return crtc;
+	if (check_crtc_for_output (crtc, output, False))
+	    return crtc;
     }
     return NULL;
 }
@@ -1378,25 +1695,40 @@ set_screen_size (void)
     {
 	XRRModeInfo *mode_info = output->mode_info;
 	int	    x, y, w, h;
+	box_t	    bounds;
 	
 	if (!mode_info) continue;
 	
-	x = output->x;
-	y = output->y;
-	w = mode_width (mode_info, output->rotation);
-	h = mode_height (mode_info, output->rotation);
+	mode_geometry (mode_info, output->rotation,
+		       &output->transform.transform,
+		       &bounds);
+	x = output->x + bounds.x1;
+	y = output->y + bounds.y1;
+	w = bounds.x2 - bounds.x1;
+	h = bounds.y2 - bounds.y1;
 	/* make sure output fits in specified size */
 	if (fb_specified)
 	{
 	    if (x + w > fb_width || y + h > fb_height)
-		fatal ("specified screen %dx%d not large enough for output %s (%dx%d+%d+%d)\n",
-		       fb_width, fb_height, output->output.string, w, h, x, y);
+		warning ("specified screen %dx%d not large enough for output %s (%dx%d+%d+%d)\n",
+			 fb_width, fb_height, output->output.string, w, h, x, y);
 	}
 	/* fit fb to output */
 	else
 	{
-	    if (x + w > fb_width) fb_width = x + w;
-	    if (y + h > fb_height) fb_height = y + h;
+	    XRRPanning *pan;
+	    if (x + w > fb_width)
+		fb_width = x + w;
+	    if (y + h > fb_height)
+		fb_height = y + h;
+	    if (output->changes & changes_panning)
+		pan = &output->panning;
+	    else
+		pan = output->crtc_info ? output->crtc_info->panning_info : NULL;
+	    if (pan && pan->left + pan->width > fb_width)
+		fb_width = pan->left + pan->width;
+	    if (pan && pan->top + pan->height > fb_height)
+		fb_height = pan->top + pan->height;
 	}
     }	
 
@@ -1417,6 +1749,130 @@ set_screen_size (void)
     
 #endif
     
+void
+disable_outputs (output_t *outputs)
+{
+    while (outputs)
+    {
+	outputs->crtc_info = NULL;
+	outputs = outputs->next;
+    }
+}
+
+/*
+ * find the best mapping from output to crtc available
+ */
+int
+pick_crtcs_score (output_t *outputs)
+{
+    output_t	*output;
+    int		best_score;
+    int		my_score;
+    int		score;
+    crtc_t	*best_crtc;
+    int		c;
+    
+    if (!outputs)
+	return 0;
+    
+    output = outputs;
+    outputs = outputs->next;
+    /*
+     * Score with this output disabled
+     */
+    output->crtc_info = NULL;
+    best_score = pick_crtcs_score (outputs);
+    if (output->mode_info == NULL)
+	return best_score;
+
+    best_crtc = NULL;
+    /* 
+     * Now score with this output any valid crtc
+     */
+    for (c = 0; c < output->output_info->ncrtc; c++)
+    {
+	crtc_t	    *crtc;
+
+	crtc = find_crtc_by_xid (output->output_info->crtcs[c]);
+	if (!crtc)
+	    fatal ("cannot find crtc 0x%x\n", output->output_info->crtcs[c]);
+	
+	/* reset crtc allocation for following outputs */
+	disable_outputs (outputs);
+	if (!check_crtc_for_output (crtc, output, True))
+	    continue;
+	
+	my_score = 1000;
+	/* slight preference for existing connections */
+	if (crtc == output->current_crtc_info)
+	    my_score++;
+
+	output->crtc_info = crtc;
+	score = my_score + pick_crtcs_score (outputs);
+	if (score > best_score)
+	{
+	    best_crtc = crtc;
+	    best_score = score;
+	}
+    }
+    if (output->crtc_info != best_crtc)
+	output->crtc_info = best_crtc;
+    /*
+     * Reset other outputs based on this one using the best crtc
+     */
+    (void) pick_crtcs_score (outputs);
+
+    return best_score;
+}
+
+/*
+ * Pick crtcs for any changing outputs that don't have one
+ */
+void
+pick_crtcs (void)
+{
+    output_t	*output;
+
+    /*
+     * First try to match up newly enabled outputs with spare crtcs
+     */
+    for (output = outputs; output; output = output->next)
+    {
+	if (output->changes && output->mode_info)
+	{
+	    if (output->crtc_info) {
+		if (output->crtc_info->crtc_info->noutput > 0 &&
+		    (output->crtc_info->crtc_info->noutput > 1 ||
+		     output != find_output_by_xid (output->crtc_info->crtc_info->outputs[0])))
+		    break;
+	    } else {
+		output->crtc_info = find_crtc_for_output (output);
+		if (!output->crtc_info)
+		    break;
+	    }
+	}
+    }
+    /*
+     * Everyone is happy
+     */
+    if (!output)
+	return;
+    /*
+     * When the simple way fails, see if there is a way
+     * to swap crtcs around and make things work
+     */
+    for (output = outputs; output; output = output->next)
+	output->current_crtc_info = output->crtc_info;
+    pick_crtcs_score (outputs);
+    for (output = outputs; output; output = output->next)
+    {
+	if (output->mode_info && !output->crtc_info)
+	    fatal ("cannot find crtc for output %s\n", output->output.string);
+	if (!output->changes && output->crtc_info != output->current_crtc_info)
+	    output->changes |= changes_crtc;
+    }
+}
+
 int
 main (int argc, char **argv)
 {
@@ -1549,7 +2005,10 @@ main (int argc, char **argv)
 	    continue;
 	}
 #if HAS_RANDR_1_2
-	if (!strcmp ("--prop", argv[i]) || !strcmp ("--properties", argv[i]))
+	if (!strcmp ("--prop", argv[i]) ||
+	    !strcmp ("--props", argv[i]) ||
+	    !strcmp ("--madprops", argv[i]) ||
+	    !strcmp ("--properties", argv[i]))
 	{
 	    query_1_2 = True;
 	    properties = True;
@@ -1557,9 +2016,12 @@ main (int argc, char **argv)
 	}
 	if (!strcmp ("--output", argv[i])) {
 	    if (++i >= argc) usage();
-	    output = add_output ();
 
-	    set_name (&output->output, argv[i], name_string|name_xid);
+	    output = find_output_by_name (argv[i]);
+	    if (!output) {
+		output = add_output ();
+		set_name (&output->output, argv[i], name_string|name_xid);
+	    }
 	    
 	    setit_1_2 = True;
 	    continue;
@@ -1659,6 +2121,36 @@ main (int argc, char **argv)
 	    output->changes |= changes_relation;
 	    continue;
 	}
+	if (!strcmp ("--panning", argv[i])) {
+	    XRRPanning *pan;
+	    if (++i>=argc) usage ();
+	    if (!output) usage();
+	    pan = &output->panning;
+	    switch (sscanf (argv[i], "%dx%d+%d+%d/%dx%d+%d+%d/%d/%d/%d/%d",
+			    &pan->width, &pan->height, &pan->left, &pan->top,
+			    &pan->track_width, &pan->track_height,
+			    &pan->track_left, &pan->track_top,
+			    &pan->border_left, &pan->border_top,
+			    &pan->border_right, &pan->border_bottom)) {
+	    case 2:
+		pan->left = pan->top = 0;
+		/* fall through */
+	    case 4:
+		pan->track_left = pan->track_top =
+		    pan->track_width = pan->track_height = 0;
+		/* fall through */
+	    case 8:
+		pan->border_left = pan->border_top =
+		    pan->border_right = pan->border_bottom = 0;
+		/* fall through */
+	    case 12:
+		break;
+	    default:
+		usage ();
+	    }
+	    output->changes |= changes_panning;
+	    continue;
+	}
 	if (!strcmp ("--set", argv[i])) {
 	    output_prop_t   *prop;
 	    if (!output) usage();
@@ -1674,6 +2166,50 @@ main (int argc, char **argv)
 	    setit_1_2 = True;
 	    continue;
 	}
+	if (!strcmp ("--scale", argv[i]))
+	{
+	    double  sx, sy;
+	    if (++i>=argc) usage();
+	    if (sscanf (argv[i], "%lfx%lf", &sx, &sy) != 2)
+		usage ();
+	    init_transform (&output->transform);
+	    output->transform.transform.matrix[0][0] = XDoubleToFixed (sx);
+	    output->transform.transform.matrix[1][1] = XDoubleToFixed (sy);
+	    output->transform.transform.matrix[2][2] = XDoubleToFixed (1.0);
+	    if (sx != 1 || sy != 1)
+		output->transform.filter = "bilinear";
+	    else
+		output->transform.filter = "nearest";
+	    output->transform.nparams = 0;
+	    output->transform.params = NULL;
+	    output->changes |= changes_transform;
+	    continue;
+	}
+	if (!strcmp ("--transform", argv[i])) {
+	    double  transform[3][3];
+	    int	    k, l;
+	    if (++i>=argc) usage ();
+	    init_transform (&output->transform);
+	    if (strcmp (argv[i], "none") != 0)
+	    {
+		if (sscanf(argv[i], "%lf,%lf,%lf,%lf,%lf,%lf,%lf,%lf,%lf",
+			   &transform[0][0],&transform[0][1],&transform[0][2],
+			   &transform[1][0],&transform[1][1],&transform[1][2],
+			   &transform[2][0],&transform[2][1],&transform[2][2])
+		    != 9)
+		    usage ();
+		init_transform (&output->transform);
+		for (k = 0; k < 3; k++)
+		    for (l = 0; l < 3; l++) {
+			output->transform.transform.matrix[k][l] = XDoubleToFixed (transform[k][l]);
+		    }
+		output->transform.filter = "bilinear";
+		output->transform.nparams = 0;
+		output->transform.params = NULL;
+	    }
+	    output->changes |= changes_transform;
+	    continue;
+	}
 	if (!strcmp ("--off", argv[i])) {
 	    if (!output) usage();
 	    set_name_xid (&output->mode, None);
@@ -1762,21 +2298,6 @@ main (int argc, char **argv)
 	    if (sscanf (argv[i++], "%d", &m->mode.vTotal) != 1) usage();
 	    m->mode.modeFlags = 0;
 	    while (i < argc) {
-		static const struct {
-		    char	    *string;
-		    unsigned long   flag;
-		} mode_flags[] = {
-		    { "+HSync", RR_HSyncPositive },
-		    { "-HSync", RR_HSyncNegative },
-		    { "+VSync", RR_VSyncPositive },
-		    { "-VSync", RR_VSyncNegative },
-		    { "Interlace", RR_Interlace },
-		    { "DoubleScan", RR_DoubleScan },
-		    { "CSync",	    RR_CSync },
-		    { "+CSync",	    RR_CSyncPositive },
-		    { "-CSync",	    RR_CSyncNegative },
-		    { NULL,	    0 }
-		};
 		int f;
 		
 		for (f = 0; mode_flags[f].string; f++)
@@ -1868,6 +2389,8 @@ main (int argc, char **argv)
     }
     if (major > 1 || (major == 1 && minor >= 2))
 	has_1_2 = True;
+    if (major > 1 || (major == 1 && minor >= 3))
+	has_1_3 = True;
 	
     if (has_1_2 && modeit)
     {
@@ -2003,18 +2526,7 @@ main (int argc, char **argv)
 	set_positions ();
 	set_screen_size ();
 
-	/*
-	 * Pick crtcs for any changing outputs that don't have one
-	 */
-	for (output = outputs; output; output = output->next)
-	{
-	    if (output->changes && output->mode_info && !output->crtc_info)
-	    {
-		output->crtc_info = find_crtc_for_output (output);
-		if (!output->crtc_info)
-		    fatal ("cannot find crtc for output %s\n", output->output.string);
-	    }
-	}
+	pick_crtcs ();
 
 	/*
 	 * Assign outputs to crtcs
@@ -2078,6 +2590,11 @@ main (int argc, char **argv)
 	    }
 	}
 	
+	/*
+	 * Set panning
+	 */
+	set_panning ();
+	
 	/*
 	 * Now apply all of the changes
 	 */
@@ -2105,6 +2622,8 @@ main (int argc, char **argv)
 	for (output = outputs; output; output = output->next)
 	{
 	    XRROutputInfo   *output_info = output->output_info;
+	    crtc_t	    *crtc = output->crtc_info;
+	    XRRCrtcInfo	    *crtc_info = crtc ? crtc->crtc_info : NULL;
 	    XRRModeInfo	    *mode = output->mode_info;
 	    Atom	    *props;
 	    int		    j, k, nprop;
@@ -2114,10 +2633,16 @@ main (int argc, char **argv)
 	    printf ("%s %s", output_info->name, connection[output_info->connection]);
 	    if (mode)
 	    {
-		printf (" %dx%d+%d+%d",
-			mode_width (mode, output->rotation),
-			mode_height (mode, output->rotation),
-			output->x, output->y);
+		if (crtc_info) {
+		    printf (" %dx%d+%d+%d",
+			    crtc_info->width, crtc_info->height,
+			    crtc_info->x, crtc_info->y);
+		} else {
+		    printf (" %dx%d+%d+%d",
+			    mode->width, mode->height, output->x, output->y);
+		}
+		if (verbose)
+		    printf (" (0x%x)", mode->id);
 		if (output->rotation != RR_Rotate_0 || verbose)
 		{
 		    printf (" %s", 
@@ -2155,6 +2680,28 @@ main (int argc, char **argv)
 		printf (" %dmm x %dmm",
 			output_info->mm_width, output_info->mm_height);
 	    }
+
+	    if (crtc && crtc->panning_info && crtc->panning_info->width > 0)
+	    {
+		XRRPanning *pan = crtc->panning_info;
+		printf (" panning %dx%d+%d+%d",
+			pan->width, pan->height, pan->left, pan->top);
+		if ((pan->track_width    != 0 &&
+		     (pan->track_left    != pan->left		||
+		      pan->track_width   != pan->width		||
+		      pan->border_left   != 0			||
+		      pan->border_right  != 0))			||
+		    (pan->track_height   != 0 &&
+		     (pan->track_top     != pan->top		||
+		      pan->track_height  != pan->height		||
+		      pan->border_top    != 0			||
+		      pan->border_bottom != 0)))
+		    printf (" tracking %dx%d+%d+%d border %d/%d/%d/%d",
+			    pan->track_width,  pan->track_height,
+			    pan->track_left,   pan->track_top,
+			    pan->border_left,  pan->border_top,
+			    pan->border_right, pan->border_bottom);
+	    }
 	    printf ("\n");
 
 	    if (verbose)
@@ -2180,6 +2727,33 @@ main (int argc, char **argv)
 			printf (" %d", crtc->crtc.index);
 		}
 		printf ("\n");
+		if (output->crtc_info && output->crtc_info->panning_info) {
+		    XRRPanning *pan = output->crtc_info->panning_info;
+		    printf ("\tPanning:    %dx%d+%d+%d\n",
+			    pan->width, pan->height, pan->left, pan->top);
+		    printf ("\tTracking:   %dx%d+%d+%d\n",
+			    pan->track_width,  pan->track_height,
+			    pan->track_left,   pan->track_top);
+		    printf ("\tBorder:     %d/%d/%d/%d\n",
+			    pan->border_left,  pan->border_top,
+			    pan->border_right, pan->border_bottom);
+		}
+	    }
+	    if (verbose)
+	    {
+		int x, y;
+
+		printf ("\tTransform: ");
+		for (y = 0; y < 3; y++)
+		{
+		    for (x = 0; x < 3; x++)
+			printf (" %f", XFixedToDouble (output->transform.transform.matrix[y][x]));
+		    if (y < 2)
+			printf ("\n\t           ");
+		}
+		if (output->transform.filter)
+		    printf ("\n\t           filter: %s", output->transform.filter);
+		printf ("\n");
 	    }
 	    if (verbose || properties)
 	    {
@@ -2248,12 +2822,14 @@ main (int argc, char **argv)
 			    }
 			}
 			printf("\n");
-			
 		    } else if (actual_format == 8) {
 			printf ("\t\t%s: %s%s\n", XGetAtomName (dpy, props[j]),
 				prop, bytes_after ? "..." : "");
 		    } else {
-			printf ("\t\t%s: ????\n", XGetAtomName (dpy, props[j]));
+			char	*type = actual_type ? XGetAtomName (dpy, actual_type) : "none";
+			printf ("\t\t%s: %s(%d) (format %d items %d) ????\n",
+				XGetAtomName (dpy, props[j]),
+				type, actual_type, actual_format, nitems);
 		    }
 
 		    free(propinfo);
@@ -2265,10 +2841,19 @@ main (int argc, char **argv)
 		for (j = 0; j < output_info->nmode; j++)
 		{
 		    XRRModeInfo	*mode = find_mode_by_xid (output_info->modes[j]);
+		    int		f;
 		    
-		    printf ("  %s (0x%x) %6.1fMHz\n",
+		    printf ("  %s (0x%x) %6.1fMHz",
 			    mode->name, mode->id,
 			    (float)mode->dotClock / 1000000.0);
+		    for (f = 0; mode_flags[f].flag; f++)
+			if (mode->modeFlags & mode_flags[f].flag)
+			    printf (" %s", mode_flags[f].string);
+		    if (mode == output->mode_info)
+			printf (" *current");
+		    if (j < output_info->npreferred)
+			printf (" +preferred");
+		    printf ("\n");
 		    printf ("        h: width  %4d start %4d end %4d total %4d skew %4d clock %6.1fKHz\n",
 			    mode->width, mode->hSyncStart, mode->hSyncEnd,
 			    mode->hTotal, mode->hSkew, mode_hsync (mode) / 1000);
